"""
Zeraora

A personal utility package, with long time supports.
"""

VERSION = (0, 4, 0, 'alpha', 1)

# https://packaging.python.org/en/latest/specifications/version-specifiers/
__version__ = '0.4.0a1'
